@@ -77,7 +77,7 @@ class StrMixin:
             value = getattr(self, name)
             if value is None:
                 continue
-            if isinstance(value, ctypes.Array):
+            if isinstance(value, (list, tuple, ctypes.Array)):
                 if len(value) == 0:
                     value = '[]'
                 elif isinstance(value[0], (int, float, str, ctypes._SimpleCData)):
@@ -258,25 +258,6 @@ class NV_GPU_PERF_PSTATES20_PARAM_DELTA(StrStructure):
                 ('valueMin', ctypes.c_int32),
                 ('valueMax', ctypes.c_int32)]
 
-class _NV_GPU_PSTATE_DATA_RANGE(StrStructure):
-    _fields_ = [('minFreq_kHz', ctypes.c_uint32),
-                ('maxFreq_kHz', ctypes.c_uint32),
-                ('domainId', ctypes.c_int),
-                ('minVoltage_uV', ctypes.c_uint32),
-                ('maxVoltage_uV', ctypes.c_uint32)]
-
-class _NV_GPU_PSTATE_DATA_U(StrUnion):
-    _fields_ = [('single_freq_kHz', ctypes.c_uint32),
-                ('range', _NV_GPU_PSTATE_DATA_RANGE)]
-
-class NV_GPU_PSTATE20_CLOCK_ENTRY_V1(StrStructure):
-    _fields_ = [('domainId', ctypes.c_int),
-                ('typeId', ctypes.c_int),
-                ('bIsEditable', ctypes.c_uint32, 1),
-                ('reserved', ctypes.c_uint32, 31),
-                ('freqDelta_kHz', NV_GPU_PERF_PSTATES20_PARAM_DELTA),
-                ('data', _NV_GPU_PSTATE_DATA_U)]
-
 class NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1(StrStructure):
     _fields_ = [('domainId', ctypes.c_int),
                 ('bIsEditable', ctypes.c_uint32, 1),
@@ -284,18 +265,93 @@ class NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1(StrStructure):
                 ('volt_uV', ctypes.c_uint32),
                 ('voltDelta_uV', NV_GPU_PERF_PSTATES20_PARAM_DELTA)]
 
-class _NV_GPU_PSTATE(StrStructure):
-    _fields_ = [('pstateId', ctypes.c_int),
-                ('bIsEditable', ctypes.c_uint32, 1),
-                ('reserved', ctypes.c_uint32, 31),
-                ('clocks', NV_GPU_PSTATE20_CLOCK_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_CLOCKS),
-                ('baseVoltages', NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES)]
+class ClockType(enum.IntEnum):
+    SINGLE = 0
+    RANGE = 1
 
-class _NV_GPU_OVERVOLT(StrStructure):
-    _fields_ = [('numVoltages', ctypes.c_uint32),
-                ('voltages', NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES)]
+class NV_GPU_PUBLIC_CLOCK_ID(enum.IntEnum):
+    GRAPHICS = 0
+    MEMORY = 1
+    PROCESSOR = 2
+    VIDEO = 3
+    MEMORY2 = 4
+    UNDEFINED = 5
 
 class NV_GPU_PERF_PSTATES20_INFO(NvVersioned):
+    class _NV_GPU_PSTATE(StrStructure):
+        class NV_GPU_PSTATE20_CLOCK_ENTRY_V1(StrStructure):
+            class _NV_GPU_PSTATE_DATA_U(StrUnion):
+                class _NV_GPU_PSTATE_DATA_RANGE(StrStructure):
+                    _fields_ = [('_minFreq', ctypes.c_uint32),
+                                ('_maxFreq', ctypes.c_uint32),
+                                ('domainId', ctypes.c_int),
+                                ('_minVoltage', ctypes.c_uint32),
+                                ('_maxVoltage', ctypes.c_uint32)]
+                    @property
+                    def minFreq(self):
+                        return self._minFreq / 1000.0
+                    @property
+                    def maxFreq(self):
+                        return self._maxFreq / 1000.0
+                    @property
+                    def minVoltage(self):
+                        return self._minVoltage / 1000000.0
+                    @property
+                    def maxVoltage(self):
+                        return self._maxVoltage / 1000000.0
+
+                _fields_ = [('_singleFreq', ctypes.c_uint32),
+                            ('range', _NV_GPU_PSTATE_DATA_RANGE)]
+                @property
+                def singleFreq(self):
+                    return self._singleFreq / 1000.0
+
+            _fields_ = [('_domainId', ctypes.c_int),
+                        ('_typeId', ctypes.c_int),
+                        ('bIsEditable', ctypes.c_uint32, 1),
+                        ('reserved', ctypes.c_uint32, 31),
+                        ('freqDelta_kHz', NV_GPU_PERF_PSTATES20_PARAM_DELTA),
+                        ('_data', _NV_GPU_PSTATE_DATA_U)]
+            @property
+            def domainId(self):
+                return NV_GPU_PUBLIC_CLOCK_ID(self._domainId)
+            @property
+            def typeId(self):
+                return ClockType(self._typeId)
+            @property
+            def data(self):
+                attr = {ClockType.SINGLE: 'singleFreq',
+                        ClockType.RANGE: 'range'}[self.typeId]
+                return getattr(self._data, attr)
+
+        _fields_ = [('pstateId', ctypes.c_int),
+                    ('bIsEditable', ctypes.c_uint32, 1),
+                    ('reserved', ctypes.c_uint32, 31),
+                    ('_clocks', NV_GPU_PSTATE20_CLOCK_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_CLOCKS),
+                    ('_baseVoltages', NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES)]
+        def _get_nums_from_parent_pstate20(self):
+            pstates_array = self._b_base_
+            if pstates_array is None:
+                return NVAPI_MAX_GPU_PSTATE20_CLOCKS, NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES
+            pstate20 = pstates_array._b_base_
+            if pstate20 is None or not isinstance(pstate20, NV_GPU_PERF_PSTATES20_INFO):
+                return NVAPI_MAX_GPU_PSTATE20_CLOCKS, NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES
+            return pstate20.numClocks, pstate20.numBaseVoltages
+
+        @property
+        def clocks(self):
+            return self._clocks[:self._get_nums_from_parent_pstate20()[0]]
+        @property
+        def baseVoltages(self):
+            return self._baseVoltages[:self._get_nums_from_parent_pstate20()[1]]
+
+    class _NV_GPU_OVERVOLT(StrStructure):
+        _fields_ = [('numVoltages', ctypes.c_uint32),
+                    ('_voltages', NV_GPU_PSTATE20_BASE_VOLTAGE_ENTRY_V1 * NVAPI_MAX_GPU_PSTATE20_BASE_VOLTAGES)]
+        @property
+        def voltages(self):
+            return self._voltages[:self.numVoltages]
+
     _nv_version_ = 2
     _fields_ = [('version', ctypes.c_uint32),
                 ('bIsEditable', ctypes.c_uint32, 1),
@@ -303,8 +359,11 @@ class NV_GPU_PERF_PSTATES20_INFO(NvVersioned):
                 ('numPstates', ctypes.c_uint32),
                 ('numClocks', ctypes.c_uint32),
                 ('numBaseVoltages', ctypes.c_uint32),
-                ('pstates', _NV_GPU_PSTATE * NVAPI_MAX_GPU_PSTATE20_PSTATES),
+                ('_pstates', _NV_GPU_PSTATE * NVAPI_MAX_GPU_PSTATE20_PSTATES),
                 ('ov', _NV_GPU_OVERVOLT)]
+    @property
+    def pstates(self):
+        return self._pstates[:self.numPstates]
 
 class _NV_GPU_POWER_INFO_ENTRY(StrStructure):
     _fields_ = [('pstate', ctypes.c_uint32),
@@ -351,6 +410,65 @@ class NV_GPU_TOPOLOGY_STATUS(NvVersioned):
                 ('count', ctypes.c_uint32),
                 ('entries', NV_GPU_TOPOLOGY_ENTRY * NVAPI_MAX_GPU_TOPOLOGY_ENTRIES)]
 
+class PowerChannelType(enum.IntEnum):
+    DEFAULT = 0
+    SUMMATION = 1
+    ESTIMATION = 2
+    SLOW = 3
+    GEMINI_CORRECTION = 4
+    C1X = 5
+    SENSOR = 6
+    PSTATE_ESTIMATION_LUT = 7
+    SENSOR_CLIENT_ALIGNED = 8
+
+class PowerRailType(enum.IntEnum):
+    UNKNOWN = 0
+    OUT_NVVDD = 1
+    OUT_FBVDD = 2
+    OUT_FBVDDQ = 3
+    OUT_FBVDD_Q = 4
+    OUT_PEXVDD = 5
+    OUT_A3V3 = 6
+    OUT_3V3NV = 7
+    OUT_TOTAL_GPU = 8
+    OUT_FBVDDQ_GPU = 9
+    OUT_FBVDDQ_MEM = 10
+    OUT_SRAM = 11
+    IN_PEX12V1 = 222
+    IN_TOTAL_BOARD2 = 223
+    IN_HIGH_VOLT0 = 224
+    IN_HIGH_VOLT1 = 225
+    IN_NVVDD1 = 226
+    IN_NVVDD2 = 227
+    IN_EXT12V_8PIN2 = 228
+    IN_EXT12V_8PIN3 = 229
+    IN_EXT12V_8PIN4 = 230
+    IN_EXT12V_8PIN5 = 231
+    IN_MISC0 = 232
+    IN_MISC1 = 233
+    IN_MISC2 = 234
+    IN_MISC3 = 235
+    IN_USBC0 = 236
+    IN_USBC1 = 237
+    IN_FAN0 = 238
+    IN_FAN1 = 239
+    IN_SRAM = 240
+    IN_PWR_SRC_PP = 241
+    IN_3V3_PP = 242
+    IN_3V3_MAIN = 243
+    IN_3V3_AON = 244
+    IN_TOTAL_BOARD = 245
+    IN_NVVDD = 246
+    IN_FBVDD = 247
+    IN_FBVDDQ = 248
+    IN_FBVDD_Q = 249
+    IN_EXT12V_8PIN0 = 250
+    IN_EXT12V_8PIN1 = 251
+    IN_EXT12V_6PIN0 = 252
+    IN_EXT12V_6PIN1 = 253
+    IN_PEX3V3 = 254
+    IN_PEX12V = 255
+
 class NV_POWER_MONITOR_INFO(NvVersioned):
     class NV_POWER_MONITOR_INFO_CHANNEL_INFO(StrStructure):
         class NV_POWER_MONITOR_INFO_CHANNEL_INFO_DATA(StrUnion):
@@ -392,73 +510,23 @@ class NV_POWER_MONITOR_INFO(NvVersioned):
                     ('reserved', ctypes.c_uint8 * 8),
                     ('_data', NV_POWER_MONITOR_INFO_CHANNEL_INFO_DATA)]
         TYPES = {
-            0: ('DEFAULT', None),
-            1: ('SUMMATION', 'sum'),
-            2: ('ESTIMATION', None),
-            3: ('SLOW', 'slow'),
-            4: ('GEMINI_CORRECTION', 'gemmCorr'),
-            5: ('1X', 'c1x'),
-            6: ('SENSOR', 'sensor'),
-            7: ('PSTATE_ESTIMATION_LUT', 'pstateEstLUT'),
-            8: ('SENSOR_CLIENT_ALIGNED', None)
-        }
-        RAILS = {
-            0: 'UNKNOWN',
-            1: 'OUT_NVVDD',
-            2: 'OUT_FBVDD',
-            3: 'OUT_FBVDDQ',
-            4: 'OUT_FBVDD_Q',
-            5: 'OUT_PEXVDD',
-            6: 'OUT_A3V3',
-            7: 'OUT_3V3NV',
-            8: 'OUT_TOTAL_GPU',
-            9: 'OUT_FBVDDQ_GPU',
-            10: 'OUT_FBVDDQ_MEM',
-            11: 'OUT_SRAM',
-            222: 'IN_PEX12V1',
-            223: 'IN_TOTAL_BOARD2',
-            224: 'IN_HIGH_VOLT0',
-            225: 'IN_HIGH_VOLT1',
-            226: 'IN_NVVDD1',
-            227: 'IN_NVVDD2',
-            228: 'IN_EXT12V_8PIN2',
-            229: 'IN_EXT12V_8PIN3',
-            230: 'IN_EXT12V_8PIN4',
-            231: 'IN_EXT12V_8PIN5',
-            232: 'IN_MISC0',
-            233: 'IN_MISC1',
-            234: 'IN_MISC2',
-            235: 'IN_MISC3',
-            236: 'IN_USBC0',
-            237: 'IN_USBC1',
-            238: 'IN_FAN0',
-            239: 'IN_FAN1',
-            240: 'IN_SRAM',
-            241: 'IN_PWR_SRC_PP',
-            242: 'IN_3V3_PP',
-            243: 'IN_3V3_MAIN',
-            244: 'IN_3V3_AON',
-            245: 'IN_TOTAL_BOARD',
-            246: 'IN_NVVDD',
-            247: 'IN_FBVDD',
-            248: 'IN_FBVDDQ',
-            249: 'IN_FBVDD_Q',
-            250: 'IN_EXT12V_8PIN0',
-            251: 'IN_EXT12V_8PIN1',
-            252: 'IN_EXT12V_6PIN0',
-            253: 'IN_EXT12V_6PIN1',
-            254: 'IN_PEX3V3',
-            255: 'IN_PEX12V'
+            PowerChannelType.SUMMATION: 'sum',
+            PowerChannelType.SLOW: 'slow',
+            PowerChannelType.GEMINI_CORRECTION: 'gemmCorr',
+            PowerChannelType.C1X: 'c1x',
+            PowerChannelType.SENSOR: 'sensor',
+            PowerChannelType.PSTATE_ESTIMATION_LUT: 'pstateEstLUT',
         }
+
         @property
         def rail(self):
-            return self.RAILS.get(self._rail, str(self._rail))
+            return PowerRailType(self._rail)
         @property
         def type(self):
-            return self.TYPES.get(self._type, [str(self._type)])[0]
+            return PowerChannelType(self._type)
         @property
         def data(self):
-            attr = self.TYPES.get(self._type, [None, None])[1]
+            attr = self.TYPES.get(self.type)
             if attr:
                 return getattr(self._data, attr)
             return None
@@ -571,6 +639,113 @@ class NV_POWER_MONITOR_STATUS(NvVersioned):
     def totalPower(self) -> float:
         return self._totalPower / 1000.0
 
+class NV_GPU_VOLTAGE_STATUS(NvVersioned):
+    _nv_version_ = 1
+    _fields_ = [('version', ctypes.c_uint32),
+                ('reserved1', ctypes.c_uint32),
+                ('reserved2', ctypes.c_uint32 * 8),
+                ('_voltage', ctypes.c_uint32),
+                ('reserved3', ctypes.c_uint32 * 8)]
+    @property
+    def voltage(self):
+        return self._voltage / 1000000.0
+
+class NV_GPU_CLOCKBOOST_MASK(NvVersioned):
+    class NV_GPU_CLOCKBOOST_MASK_CLOCK(StrStructure):
+        _fields_ = [('_type', ctypes.c_uint32),
+                    ('enabled', ctypes.c_bool),
+                    ('reserved', ctypes.c_uint32 * 4)]
+        @property
+        def type(self):
+            return NV_GPU_PUBLIC_CLOCK_ID(self._type)
+
+    _nv_version_ = 1
+    _pack_ = 8
+    _fields_ = [('version', ctypes.c_uint32),
+                ('masks', ctypes.c_uint32 * 16),
+                ('clocks', NV_GPU_CLOCKBOOST_MASK_CLOCK * 255)]
+
+class NV_GPU_CLOCKBOOST_TABLE(NvVersioned):
+    class NV_GPU_CLOCKBOOST_TABLE_CLOCK(StrStructure):
+        _fields_ = [('_type', ctypes.c_uint32),
+                    ('reserved1', ctypes.c_uint32 * 4),
+                    ('_freqDelta', ctypes.c_int32),
+                    ('reserved2', ctypes.c_uint32 * 3)]
+        @property
+        def freqDelta(self):
+            return self._freqDelta / 1000.0
+        @property
+        def type(self):
+            return NV_GPU_PUBLIC_CLOCK_ID(self._type)
+
+    _nv_version_ = 1
+    _pack_ = 8
+    _fields_ = [('version', ctypes.c_uint32),
+                ('masks', ctypes.c_uint32 * 16),
+                ('clocks', NV_GPU_CLOCKBOOST_TABLE_CLOCK * 255)]
+
+class NV_GPU_VFP_CURVE(NvVersioned):
+    class NV_GPU_VFP_CURVE_CLOCK(StrStructure):
+        _fields_ = [('_type', ctypes.c_uint32),
+                    ('_frequency', ctypes.c_uint32),
+                    ('_voltage', ctypes.c_uint32),
+                    ('reserved', ctypes.c_uint32 * 4)]
+        @property
+        def frequency(self):
+            return self._frequency / 1000.0
+        @property
+        def voltage(self):
+            return self._voltage / 1000000.0
+        @property
+        def type(self):
+            return NV_GPU_PUBLIC_CLOCK_ID(self._type)
+
+    _nv_version_ = 1
+    _fields_ = [('version', ctypes.c_uint32),
+                ('masks', ctypes.c_uint32 * 16),
+                ('clocks', NV_GPU_VFP_CURVE_CLOCK * 255)]
+
+class PerfCapReason(enum.IntFlag):
+    NONE = 0
+    POWER = 1
+    TEMPERATURE = 2
+    VOLTAGE = 4
+    UNKNOWN = 8
+    NO_LOAD = 16
+
+class NV_GPU_PERFORMANCE_STATUS(NvVersioned):
+    _nv_version_ = 1
+    _pack_ = 8
+    _fields_ = [('version', ctypes.c_uint32),
+                ('unknown1', ctypes.c_uint32),
+                ('_timer', ctypes.c_ulonglong),
+                ('_limit', ctypes.c_uint32),
+                ('unknown2', ctypes.c_uint32 * 3),
+                ('_timers', ctypes.c_ulonglong * 3),
+                ('unknown3', ctypes.c_uint32 * 326)]
+    @property
+    def timer(self):
+        return self._timer / 1e9
+    @property
+    def limit(self):
+        return PerfCapReason(self._limit)
+
+class RamType(enum.IntEnum):
+    Unknown = 0
+    SDRAM = 1
+    DDR1 = 2
+    DDR2 = 3
+    GDDR2 = 4
+    GDDR3 = 5
+    GDDR4 = 6
+    DDR3 = 7
+    GDDR5 = 8
+    LPDDR2 = 9
+    GDDR5X = 10
+    HBM2 = 12
+    GDDR6 = 14
+    GDDR6X = 15
+
 class Method:
     def __init__(self, offset, restype, *argtypes):
         self.proto = ctypes.CFUNCTYPE(restype, *argtypes, use_errno=True, use_last_error=True)
@@ -628,6 +803,16 @@ class NvAPI:
     NvAPI_GPU_ClientFanCoolersGetControl = NvMethod(0x814B209F, 'NvAPI_GPU_ClientFanCoolersGetControl', NvPhysicalGpu, ctypes.POINTER(NV_GPU_FAN_COOLERS_CONTROL))
     NvAPI_GPU_ClientFanCoolersSetControl = NvMethod(0xA58971A5, 'NvAPI_GPU_ClientFanCoolersSetControl', NvPhysicalGpu, ctypes.POINTER(NV_GPU_FAN_COOLERS_CONTROL))
 
+    NvAPI_GPU_GetCurrentVoltage = NvMethod(0x465F9BCF, 'NvAPI_GPU_GetCurrentVoltage', NvPhysicalGpu, ctypes.POINTER(NV_GPU_VOLTAGE_STATUS))
+
+    NvAPI_GPU_GetClockBoostMask = NvMethod(0x507B4B59, 'NvAPI_GPU_GetClockBoostMask', NvPhysicalGpu, ctypes.POINTER(NV_GPU_CLOCKBOOST_MASK))
+    NvAPI_GPU_GetVFPCurve = NvMethod(0x21537AD4, 'NvAPI_GPU_GetVFPCurve', NvPhysicalGpu, ctypes.POINTER(NV_GPU_VFP_CURVE))
+    NvAPI_GPU_GetClockBoostTable = NvMethod(0x23F1B133, 'NvAPI_GPU_GetClockBoostTable', NvPhysicalGpu, ctypes.POINTER(NV_GPU_CLOCKBOOST_TABLE))
+    NvAPI_GPU_SetClockBoostTable = NvMethod(0x733E009, 'NvAPI_GPU_SetClockBoostTable', NvPhysicalGpu, ctypes.POINTER(NV_GPU_CLOCKBOOST_TABLE))
+    NvAPI_GPU_PerfPoliciesGetStatus = NvMethod(0x3D358A0C, 'NvAPI_GPU_PerfPoliciesGetStatus', NvPhysicalGpu, ctypes.POINTER(NV_GPU_PERFORMANCE_STATUS))
+
+    NvAPI_GPU_GetRamType = NvMethod(0x57F7CAAC, 'NvAPI_GPU_GetRamType', NvPhysicalGpu, ctypes.POINTER(ctypes.c_uint32))
+
     def __init__(self):
         self.NvAPI_Initialize()
         self.__gpus = None
@@ -763,3 +948,37 @@ class NvAPI:
         if self.__version < 0x9C40:
             raise ValueError('This feature requires new drivers')
         self.NvAPI_GPU_ClientFanCoolersSetControl(dev, ctypes.pointer(control))
+
+    def get_core_voltage(self, dev: NvPhysicalGpu) -> float:
+        value = NV_GPU_VOLTAGE_STATUS()
+        self.NvAPI_GPU_GetCurrentVoltage(dev, ctypes.pointer(value))
+        return value.voltage
+
+    def get_boost_mask(self, dev: NvPhysicalGpu) -> NV_GPU_CLOCKBOOST_MASK:
+        value = NV_GPU_CLOCKBOOST_MASK()
+        self.NvAPI_GPU_GetClockBoostMask(dev, ctypes.pointer(value))
+        return value
+
+    def get_boost_table(self, dev: NvPhysicalGpu, boost_mask: NV_GPU_CLOCKBOOST_MASK) -> NV_GPU_CLOCKBOOST_TABLE:
+        value = NV_GPU_CLOCKBOOST_TABLE()
+        for i, m in enumerate(boost_mask.masks):
+            value.masks[i] = m
+        self.NvAPI_GPU_GetClockBoostTable(dev, ctypes.pointer(value))
+        return value
+
+    def get_vfp_curve(self, dev: NvPhysicalGpu, boost_mask: NV_GPU_CLOCKBOOST_MASK) -> NV_GPU_VFP_CURVE:
+        value = NV_GPU_VFP_CURVE()
+        for i, m in enumerate(boost_mask.masks):
+            value.masks[i] = m
+        self.NvAPI_GPU_GetVFPCurve(dev, ctypes.pointer(value))
+        return value
+
+    def get_performance_limit(self, dev: NvPhysicalGpu) -> PerfCapReason:
+        value = NV_GPU_PERFORMANCE_STATUS()
+        self.NvAPI_GPU_PerfPoliciesGetStatus(dev, ctypes.pointer(value))
+        return value.limit
+
+    def get_ram_type(self, dev: NvPhysicalGpu) -> RamType:
+        value = ctypes.c_uint32(0)
+        self.NvAPI_GPU_GetRamType(dev, ctypes.pointer(value))
+        return RamType(value.value)
